"""
Segments a body from a CT scan.
"""
import numpy as np
import skimage
import skimage.measure
import skimage.segmentation

import lcat


def get_body_segmentation(scan):
    """
    Given a `Scan` object representing a chest CT scan, return a binary mask representing the region
    occupied by the body.
    """
    # Threshold the image (threshold is lower limit for lung tissue in HU)
    foreground = scan.voxels >= -700

    # Identify strongly connected components
    labels = skimage.measure.label(foreground, connectivity=1)

    # Identify the largest volume
    body_mask = get_largest_volume(labels)

    # Obtain body envelope
    envelope_mask = get_body_envelope(body_mask)

    return envelope_mask


def get_largest_volume(labels):
    """
    Return a binary mask equivalent to the component with the largest volumes in the array `labels`.
    """
    return labels == get_top_value(labels[labels != 0])


def get_top_value(arr):
    """
    Given an ndarray, return the value which occurs most frequently.
    """
    # Count values
    values, counts = np.unique(arr, return_counts=True)

    # Identify top value
    top_value_index = np.argmax(counts)

    return values[top_value_index]


def get_body_envelope(body_mask):
    """
    Given a mask representing thresholded lung values, obtain an envelope containing the lung region
    with no interior holes.
    """
    # Invert the mask
    reversed_mask = np.logical_not(body_mask)

    # Identify connected_components
    reversed_labels = skimage.measure.label(reversed_mask)

    # Identify inner labels only (on x and y edges, z outer labels can remain)
    inner_labels = lcat.util.clear_border(reversed_labels, axis=[0, 1])

    # Obtain only the outermost (edge-touching) region
    outside_mask = np.logical_xor(reversed_labels != 0, inner_labels != 0)

    # Invert the mask again to get the envelope
    envelope_mask = np.logical_not(outside_mask)

    return envelope_mask



#!/usr/bin/env python
"""
Generate frames for a tracheal distance algorithm video.
"""
from __future__ import division
import argparse
import os

import numpy as np
import scipy.ndimage
from tqdm import tqdm

import lcat


DESCRIPTION = "Visualize lcat-toolkit tracheal distance calculation code."

TRACHEAL_DISTANCE_FRAME_COUNT = 180
TRACHEAL_DISTANCE_RELATIVE_WINDOW = 2 * 1 / TRACHEAL_DISTANCE_FRAME_COUNT


def generate_movie(scan_folder, destination_folder):
    """
    Generate a movie visualizing tracheal distance calculation code.
    """
    # Load scan
    print("Loading scan...")
    scan = lcat.load_scan(scan_folder, cubify=False)

    # Write frames
    write_frames(scan, destination_folder)


def write_frames(scan, destination_folder):
    """
    Write frames of tracheal distance visualization movie.
    """
    # TODO: Write original scan frame(s)

    # Perform lung segmentation
    print("Segmenting lung...")
    lung_segmentation = lcat.get_lung_segmentation(scan)

    # TODO: Write lung segmentation frame(s)

    # Get tracheal distances
    print("Calculating tracheal distances...")
    tracheal_distances = lcat.get_tracheal_distances(scan, lung_segmentation)

    # Write tracheal distance frames
    write_tracheal_distance_frames(tracheal_distances, destination_folder)


def write_tracheal_distance_frames(tracheal_distances, destination_folder):
    """
    Write tracheal distances frame to the destination folder.
    """
    # Separate mask and data
    value_mask = np.logical_not(tracheal_distances.mask)
    tracheal_distances = tracheal_distances.data

    # Identify caclulated (non-zero) pixels
    distance_mask = np.logical_and(np.logical_not(np.isclose(tracheal_distances, 0)), value_mask)

    # Identify limits
    min_distance = np.min(tracheal_distances[distance_mask])
    max_distance = np.max(tracheal_distances[distance_mask])
    distance_range = max_distance - min_distance

    # Generate frame distances
    frame_distances = np.linspace(min_distance, max_distance, TRACHEAL_DISTANCE_FRAME_COUNT)

    # Loop over each frame
    for frame, distance in enumerate(tqdm(frame_distances.flat, desc="Writing frames")):
        # Calculate frame window
        absolute_window = distance_range * TRACHEAL_DISTANCE_RELATIVE_WINDOW
        window_start = max(min_distance, distance - absolute_window / 2)
        window_end = min(max_distance, distance + absolute_window / 2)

        # Generate mask
        frame_mask = np.logical_and(window_start <= tracheal_distances,
                                    tracheal_distances <= window_end)

        # Flatten mask
        frame_mask = np.max(frame_mask, axis=2).T

        # Determine output path
        filename = 'td-%03d.png' % frame
        filepath = os.path.join(destination_folder, filename)

        # Output frame
        lcat.util.image_from_mask(frame_mask).save(filepath)


def main():
    """
    Launch tracheal distance movie generator.
    """
    # Set up arguments
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument('scan_folder', metavar="scan-folder",
                        help="Folder containing input dicom files of chest CT scans.")
    parser.add_argument('destination_folder', metavar="destination-folder",
                        help="Destination folder for slice .tiff files.")

    # Parse arguments
    args = parser.parse_args()

    # Generate tracheal distance movie
    generate_movie(args.scan_folder, args.destination_folder)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Featurize tumor tracheal distance.
"""
from __future__ import print_function
import argparse
import itertools
import os
import re

import pandas as pd
from tqdm import tqdm

import lcat
import lcat.featurization


DESCRIPTION = "Featurize tumor tracheal distance."

PATIENT_FOLDER_RE = re.compile('LIDC-IDRI-(.+)')


def execute(data_folder, destination_file):
    """
    Featurize tumor tracheal distance for all patients in `data_folder`, and write featurization to
    `destination_file`.
    """
    # Create featurizations placeholder
    featurizations = []

    # List all patient paths
    patient_paths = list(generate_patient_paths(data_folder))

    # For each patient in the folder
    for patient_id, path in tqdm(patient_paths, desc="Featurizing"):
        # Update progress bar
        tqdm.write("Loading scan for patient %s..." % patient_id)

        # Load scan
        scan = lcat.load_scan(path, cubify=True)

        tqdm.write("Featurizing scan...")

        # Featurize scan
        featurization = lcat.featurization.featurize_scan(scan)

        if len(featurization) == 0:
            continue

        # Generate new index content
        index_content = zip(itertools.repeat(scan.patient_id), featurization.index)

        # Add the patient id column
        index = pd.MultiIndex.from_tuples(index_content, names=['patient_id', 'nodule_id'])

        # Apply new index
        featurization.index = index

        # Store featurization for scan
        featurizations.append(featurization)

    # Combine featurizations
    data = pd.concat(featurizations)

    # Write data to file
    data.to_csv(destination_file)


def generate_patient_paths(data_folder):
    """
    Yield a patient ID and path for each patient folder in data_folder.
    """
    # For every filename in data_folder
    for filename in os.listdir(data_folder):
        # Generate full path
        filepath = os.path.join(data_folder, filename)

        # Make sure it's a folder
        if not os.path.isdir(filepath):
            continue

        # Perform regex matching
        match = PATIENT_FOLDER_RE.match(filename)

        # Reject improperly named folders
        if match is None:
            continue

        # Yield patient ID and path
        yield match.group(1), filepath


def main():
    """
    Launch bronchi segmentation test harness.
    """
    # Set up arguments
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument('data_folder', metavar="scan-folder",
                        help="Folder containing LIDC-IDRI data.")
    parser.add_argument('destination_file', metavar="destination-file",
                        help="Destination CSV file for featurization.")

    # Parse arguments
    args = parser.parse_args()

    # Test bronchi segmentation code
    execute(args.data_folder, args.destination_file)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
BMI 260: Final Project
Bronchi segmentation/identification
"""
import numpy as np
import skfmm
import skimage.morphology


def get_tracheal_distances(scan, lung_segmentation):
    # Obtain median threshold
    lung_threshold = np.median(scan.voxels[lung_segmentation])

    # Select non-air elements
    lung_tissue_mask = np.logical_and(lung_segmentation, scan.voxels > lung_threshold)

    # Compute boundaries affecting lung air
    air_boundaries = np.logical_or(lung_tissue_mask, np.logical_not(lung_segmentation))

    # Choose seed point
    # TODO: Choose seed point better (actually find trachea, don't assume top)
    seed_boundary = get_seed_boundary(lung_segmentation)

    # Set up fast marching problem
    phi = np.ma.MaskedArray(seed_boundary, air_boundaries)

    # Perform fast marching problem
    distances = skfmm.distance(phi)

    # TODO: Return something else?
    return distances


def get_seed_boundary(lung_segmentation):
    """
    Given a `lung_segmentation`, identify and return the top slice of the trachea.
    """
    # Trachea disc placeholder
    trachea_index = None

    # For each slice
    for z_index in range(lung_segmentation.shape[2]):
        # Set candidate trachea disc
        candidate_trachea_disc = lung_segmentation[..., z_index]

        # Check for content
        if np.sum(candidate_trachea_disc) > 0:
            trachea_index = z_index
            break

    # Make sure there was a slice with content
    if trachea_index is None:
        raise Exception("No content in lung segmentation")

    # Create same-size array of ones
    trachea_ceiling = np.ones(lung_segmentation.shape)

    # Set top of trachea as starting point
    trachea_ceiling[lung_segmentation[..., trachea_index]] = 0

    return trachea_ceiling



"""
Load data for a chest CT scan from both dicom and radiologist xml files.
"""
from __future__ import division, print_function
from collections import namedtuple
import math
import warnings

import numpy as np
import scipy.ndimage

import lcat.loading.annotations
import lcat.loading.images


# Scan datatype
Scan = namedtuple('Scan', ['patient_id', 'voxels', 'nodules', 'unit_cell'])


def load_scan(scan_folder, cubify=False):
    """
    Loads the CT scan as a 3d voxel array, then loads the segmentation in the given dicom_folder by
    reading any xml files located in the folder and referencing dicom files as necessary. Returns a
    3D mask with the same dimensions as the CT scan.

    TODO: Combine nodules referring to the same entity. Unfortunately there is currently no unique
    ID for each nodule, meaning that multiple radiologist reads result in multiple almost-identical
    nodules in the scan metadata.
    """
    # Load the CT scan
    patient_id, voxels, unit_cell, sop_instance_uids = lcat.loading.images.load_folder(scan_folder)

    # Load all segmentations
    nodules = lcat.loading.annotations.load_radiologist_annotations(scan_folder, voxels.shape,
                                                                    sop_instance_uids)

    # Convert to scan datatype
    scan = Scan(patient_id, voxels, nodules, unit_cell)

    # Cubify if necessary
    if cubify:
        scan = cubify_scan(scan)

    return scan


def cubify_scan(scan):
    """
    Given a scan, interpolate the data to make the unit cell cubic. The dimension(s) with the
    smallest magnitude(s) in the unit cell will remain the same.
    """
    # Calculate scaling factors
    scaling_factors = get_scaling_factors(scan)

    # Calculate new unit cell size
    new_unit_cell = [step / factor for step, factor in zip(scan.unit_cell, scaling_factors)]

    # Rescale the voxels
    new_voxels = zoom_array(scan.voxels, scaling_factors)

    # Create nodules placeholder
    new_nodules = []

    # For each nodule
    for nodule in scan.nodules:
        # Perform nodule mask interpolation
        new_nodules.append(rescale_nodule(nodule, scaling_factors))

    # Return new Scan
    return Scan(scan.patient_id, new_voxels, new_nodules, new_unit_cell)


def get_scaling_factors(scan):
    """
    Returns the scaling factors for the given scan.
    """
    # Calculate the smallest step in unit_cell
    minimum_step = min(scan.unit_cell)

    # Calculate scaling factors
    scaling_factors = [step / minimum_step for step in scan.unit_cell]

    return scaling_factors


def zoom_array(array, factor, mode='nearest'):
    """
    Rescale the given `array` along each axis by `factor`. If `factor` is a sequence, each axis will
    be zoomed by its corresponding factor in `factor`.
    """
    with warnings.catch_warnings():
        # Filter warnings we can't fix
        warnings.filterwarnings("ignore", message="From scipy 0.13.0, the output shape of zoom() "
                                                  "is calculated with round() instead of int() - "
                                                  "for these inputs the size of the returned array "
                                                  "has changed.")

        # Return zoomed image
        return scipy.ndimage.zoom(array, factor, mode=mode)


def rescale_nodule(nodule, scaling_factors):
    """
    This interpolation process is trickier than the voxel interpolation process, because we're not
    only rescaling the mask, we're also moving the origin of the mask. In order to properly
    perform this interpolation, we perform the following steps:
    (1) Find the scaled origin as real numbers (not rounded)
    (2) Find the scaled anti-origin (opposite corner) as real numbers (not rounded)
    (3) Count the number of scaled cells along each dimension for mask
    (4) Convert rounded extents to original space
    (5) Perform rescaling by axis coordinates
    """
    # Load old extents
    old_starts = nodule.origin
    old_ends = [start + dim for start, dim in zip(old_starts, nodule.mask.shape)]

    # Calculate new extents
    new_starts = [start * factor for start, factor in zip(old_starts, scaling_factors)]
    new_ends = [end * factor for end, factor in zip(old_ends, scaling_factors)]

    # Calculate discrete extents
    discrete_starts = [int(math.floor(start)) for start in new_starts]
    discrete_ends = [int(math.ceil(end)) for end in new_ends]
    new_shape = [end - start for start, end in zip(discrete_starts, discrete_ends)]

    # Convert to original space
    unscaled_starts = [start / factor for start, factor in zip(discrete_starts, scaling_factors)]
    unscaled_ends = [end / factor for end, factor in zip(discrete_ends, scaling_factors)]

    # Generate interpolating points in original space
    axis_coordinates = [np.linspace(unscaled_start - start, unscaled_end - start, dim)
                        for start, unscaled_start, unscaled_end, dim
                        in zip(old_starts, unscaled_starts, unscaled_ends, new_shape)]

    # Perform interpolation
    new_mask = interpolate_array_by_axis(nodule.mask, axis_coordinates, output=float,
                                         mode='nearest')

    # Convert to binary array
    new_mask = new_mask > 0.5

    # Return rescaled nodule
    return lcat.loading.annotations.Nodule(nodule.nodule_id, nodule.characteristics,
                                           discrete_starts, new_mask)


def interpolate_array(array, new_shape, output=None, mode='nearest'):
    """
    Given an array and a target shape, perform an orthogonal transformation to the new shape using
    a spline interpolation.
    """
    # Generate mapping by axis
    axis_coordinates = [np.linspace(0, old_dim, new_dim)
                        for old_dim, new_dim in zip(array.shape, new_shape)]

    return interpolate_array_by_axis(array, axis_coordinates, output=output, mode=mode)


def interpolate_array_by_axis(array, axis_coordinates, output=None, mode='nearest'):
    # Expand axis coordinates
    coordinates = np.meshgrid(*axis_coordinates)

    # Perform interpolation
    # Note that mode only applies to input boundaries, the interpolation itself is a spline
    return scipy.ndimage.map_coordinates(array, coordinates, output=output, mode=mode)



"""
Authors: Connor Brinton and Scotty Fleming
Load a CT scan from a series of dicom files.
"""
import os

import dicom
import numpy as np


def load_folder(dicom_folder):
    """
    Given a folder of dicom files, load them and return a 3D numpy array representing the scan.
    Pixel values are converted to Houndsfield units using the rescaling slope and intercept encoded
    in each dicom file.
    """
    # Make sure folder exists
    if not os.path.isdir(dicom_folder):
        raise RuntimeError("The specified folder does not exist, or is a file.")

    # List all dicom files
    dicom_files = [filename for filename in os.listdir(dicom_folder) if filename.endswith(".dcm")]
    dicom_paths = [os.path.join(dicom_folder, filename) for filename in dicom_files]

    # Load each dicom object
    dicom_objects = [dicom.read_file(dicom_path) for dicom_path in dicom_paths]

    # Sort by instance number
    dicom_objects.sort(key=lambda x: x.InstanceNumber)

    # Get patient ID
    patient_id = get_single_value(dicom_object.PatientID for dicom_object in dicom_objects)

    # Obtain slice SOP instance UIDs
    sop_instance_uids = [dicom_object.SOPInstanceUID for dicom_object in dicom_objects]

    # Extract pixel arrays
    pixel_arrays = [dicom_object.pixel_array for dicom_object in dicom_objects]

    # Extract physical slice dimensions
    pixel_spacing = get_single_value(dicom_object.PixelSpacing for dicom_object in dicom_objects)
    x_spacing, y_spacing = float(pixel_spacing[0]), float(pixel_spacing[1])

    # Extract slice separation distance
    slice_positions = [dicom_object.ImagePositionPatient[2] for dicom_object in dicom_objects]
    slice_separations = abs(np.diff(slice_positions))
    z_spacing = get_single_value(slice_separations)

    # Collapse physical dimensions into unit cell
    unit_cell = (x_spacing, y_spacing, z_spacing)

    # Rescale pixels
    rescale_slopes = np.asarray([dicom_object.RescaleSlope for dicom_object in dicom_objects],
                                dtype=np.float32)
    rescale_intercepts = np.asarray([dicom_object.RescaleIntercept
                                     for dicom_object in dicom_objects], dtype=np.float32)
    rescaled_pixels = rescale_slopes[:, np.newaxis, np.newaxis] * np.asarray(pixel_arrays) \
                        + rescale_intercepts[:, np.newaxis, np.newaxis]

    # Roll pixels into xyz coordinate system
    voxels = np.moveaxis(rescaled_pixels, 0, -1)

    return patient_id, voxels, unit_cell, sop_instance_uids


def get_single_value(values):
    """
    Given a sequence of values, checks that all values are the same, and then returns the single
    value. The values must satisfy transitive equality.
    """
    # Attempt to obtain an iterator
    iterator = iter(values)

    # Get the initial value
    try:
        value = next(iterator)
    except StopIteration:
        raise ValueError("values cannot be of length zero.")

    # Make sure subsequent values match
    try:
        next_value = next(iterator)
        if value != next_value:
            raise ValueError("Not all elements in values are equal.")
    except StopIteration:
        pass

    # Everything matched, return value
    return value



"""
Authors: Connor Brinton and Scotty Fleming
Segment lungs from a chest CT scan.
"""
import numpy as np
import skimage
import skimage.filters
import skimage.measure
import skimage.morphology
import skimage.segmentation

import lcat


def get_lung_segmentation(scan):
    """
    Given a `Scan` object representing a chest CT scan, return a binary mask representing the lungs
    (not including the air within the lungs).
    """
    # Identify filler pixels (top edge value)
    filler_value = get_top_value(np.concatenate((scan.voxels[[0, -1], :, :].flat,
                                                 scan.voxels[:, [0, -1], :].flat)))

    # Determine threshold
    scan_mask = np.logical_not(np.isclose(scan.voxels, filler_value))
    threshold = skimage.filters.threshold_otsu(scan.voxels[scan_mask])

    # Make sure threshold is within air/lung parameters
    if threshold <= -1000 or threshold >= 0:
        threshold = -500

    # Threshold the image
    foreground = scan.voxels >= threshold

    # Identify strongly connected components
    labels = skimage.measure.label(foreground, background=1, connectivity=1)

    # Remove edge components
    lcat.util.clear_border(labels, axis=[0, 1], in_place=True)

    # Identify the largest volume
    lung_mask = get_largest_volume(labels)

    # Fill edge holes by dilation
    # smoother = skimage.morphology.ball(10, dtype=bool)
    # lung_mask = skimage.morphology.binary_dilation(lung_mask, selem=smoother)

    smoother = skimage.morphology.disk(10, dtype=bool)
    for z_index in range(lung_mask.shape[-1]):
        lung_mask[..., z_index] = skimage.morphology.binary_dilation(lung_mask[..., z_index],
                                                                     selem=smoother)

    # Obtain lung envelope
    envelope_mask = get_lung_envelope(lung_mask)

    # Erode envelope to revert to proper extent
    # envelope_mask = skimage.morphology.binary_erosion(envelope_mask, selem=smoother)
    for z_index in range(lung_mask.shape[-1]):
        lung_mask[..., z_index] = skimage.morphology.binary_erosion(envelope_mask[..., z_index],
                                                                    selem=smoother)

    return envelope_mask


def get_largest_volume(labels):
    """
    Return a binary mask equivalent to the component with the largest volumes in the array `labels`.
    """
    return labels == get_top_value(labels[labels != 0])


def get_top_value(arr):
    """
    Given an ndarray, return the value which occurs most frequently.
    """
    # Count values
    values, counts = np.unique(arr, return_counts=True)

    # Identify top value
    top_value_index = np.argmax(counts)

    return values[top_value_index]


def get_lung_envelope(lung_mask):
    """
    Given a mask representing thresholded lung values, obtain an envelope containing the lung region
    with no interior holes.
    """
    # Invert the mask
    reversed_mask = np.logical_not(lung_mask)

    # Identify connected_components
    reversed_labels = skimage.measure.label(reversed_mask)

    # Identify inner labels only
    inner_labels = lcat.util.clear_border(reversed_labels, axis=[0, 1])

    # Obtain only the outermost (edge-touching) region
    outside_mask = np.logical_xor(reversed_labels != 0, inner_labels != 0)

    # Invert the mask again to get the envelope
    envelope_mask = np.logical_not(outside_mask)

    return envelope_mask



#!/usr/bin/env python
"""
BMI 260: Final Project
Load chest CT scan annotations from radiologist xml files.
"""
from collections import namedtuple
import os
import re
import xml.etree.ElementTree as ET

import numpy as np
import skimage
import skimage.measure
import skimage.segmentation

import lcat


# Nodule datatype
Nodule = namedtuple('Nodule', ['nodule_id', 'characteristics', 'origin', 'mask'])

# XML namespace abbreviations
XMLNS = {
    'nih': 'http://www.nih.gov'
}

# Tag name regex
TAG_NAME_RE = re.compile('^{' + XMLNS['nih'] + '}' + '(.+)$')


def load_radiologist_annotations(dicom_folder, dimensions, sop_instance_uids):
    """
    Load radiologist annotations (namely nodule characteristics and regions) from the xml files
    present in `dicom_folder`. Returns an array of Nodule objects representing all nodules found in
    the radiologist annotations.
    """
    # Create nodules placeholder
    nodules = []

    # Look for XML files
    for filename in os.listdir(dicom_folder):
        if filename.endswith('.xml'):
            # Reconstruct filepath
            filepath = os.path.join(dicom_folder, filename)

            # Load xml file
            tree = ET.parse(filepath)
            root = tree.getroot()

            # Find all nodules
            reads = root.findall('.//nih:readingSession//nih:unblindedReadNodule', XMLNS)

            # For each read
            for read in reads:
                # Extract nodule information
                nodule = get_nodule_information(read, dimensions, sop_instance_uids)

                # Only include >3mm nodules
                if any(dim > 1 for dim in nodule.mask.shape):
                    nodules.append(nodule)

    return nodules


def get_nodule_information(read, dimensions, sop_instance_uids):
    """
    Given an unblindedReadNodule element, create a Nodule object representing the nodule's
    characteristics and vertices.
    """
    # Get nodule ID
    nodule_id = get_read_nodule_id(read)

    # Get characteristics
    characteristics = get_read_characteristics(read)

    # Get mask
    origin, mask = get_read_mask(read, dimensions, sop_instance_uids)

    return Nodule(nodule_id, characteristics, origin, mask)


def get_read_nodule_id(read):
    # Find nodule ID element
    nodule_id_elem = read.find('.//nih:noduleID', XMLNS)

    # Return text content
    return nodule_id_elem.text


def get_read_characteristics(read):
    """
    Get the characteristics from a read as recorded by the radiologist. Returns an empty dictionary
    if no characteristics were recorded.
    """
    # Extract characteristics
    characteristics = {}
    for attribute_elem in read.findall('.//nih:characteristics//*', XMLNS):
        # Get attribute name (removing namespace)
        match = TAG_NAME_RE.match(attribute_elem.tag)
        assert match is not None
        attribute_name = match.group(1)

        # Get attribute value
        attribute_value = int(attribute_elem.text)

        characteristics[attribute_name] = attribute_value

    return characteristics


def get_read_mask(read, dimensions, sop_instance_uids):
    """
    Get a 3D array representing the region described by the specific read, prefaced by an origin
    specifying its placement in the image (in index coordinates).
    """
    # Get the full mask
    mask = get_mask_region(read, dimensions, sop_instance_uids)

    # Compress to small region with offset
    origin, mask = lcat.util.compress_nodule_mask(mask)

    return origin, mask


def get_mask_region(read, dimensions, sop_instance_uids):
    """
    Returns a full representation of the region represented by the given nodule read as a mask.
    """
    # Create mask output placeholder
    mask = np.zeros(dimensions, dtype=bool)

    # Create holes queue
    holes = []

    # Identify regions of interest
    for roi_elem in read.findall('.//nih:roi', XMLNS):
        # Check if it's a hole
        if roi_elem.find('.//nih:inclusion', XMLNS).text.upper() == 'FALSE':
            holes.append(roi_elem)
        else:
            mark_region(mask, roi_elem, sop_instance_uids)

    # Create unincluded mask placeholder
    unincluded = np.zeros(dimensions, dtype=bool)

    # Identify hole regions
    for roi_elem in holes:
        mark_region(unincluded, roi_elem, sop_instance_uids)

    # Remove unincluded regions
    mask &= np.logical_not(unincluded)

    return mask


def mark_region(mask, roi_elem, sop_instance_uids):
    """
    Mark the region of interest encoded by `roi_elem` in `mask`. `sop_instance_uids` is used to
    determine the slices referenced by `roi_elem`.
    """
    # Create mask boundary placeholder
    mask_boundary = np.zeros(mask.shape[:2], dtype=bool)

    # Get Z index
    sop_instance_uid = roi_elem.find('.//nih:imageSOP_UID', XMLNS).text
    z_index = sop_instance_uids.index(sop_instance_uid)

    # Mark boundary points
    for edge_elem in roi_elem.findall('.//nih:edgeMap', XMLNS):
        # Get x and y positions
        x_position = int(edge_elem.find('.//nih:xCoord', XMLNS).text)
        y_position = int(edge_elem.find('.//nih:yCoord', XMLNS).text)

        # Mark boundary in mask
        mask_boundary[x_position, y_position] = 1

    # Fill in region
    mask_regions = skimage.measure.label(mask_boundary, background=-1, connectivity=1)
    mask_center = skimage.segmentation.clear_border(mask_regions)
    mask[:, :, z_index] |= mask_center != 0


def main():
    """
    Command-line invocation routine.
    """
    import scans
    scan = scans.load_scan('../../data/LIDC-IDRI/LIDC-IDRI-0090')

    import IPython
    IPython.embed()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Setup file for lcat-toolkit.
"""
from setuptools import setup

PACKAGES = ['lcat',
            'lcat.analysis',
            'lcat.loading',
            'lcat.segmentation']

PACKAGE_DIR = {}

PACKAGE_DATA = {}

SCRIPTS = ['scripts/lcat-featurize.py',
           'scripts/lcat-visualize.py']

setup(name='lcat',
      packages=PACKAGES,
      version='0.1.0',
      description='Lung Cancer Action Team (LCAT) toolkit',
      author='LCAT Team',
      author_email='brinton@cs.stanford.edu',
      url='https://github.com/connorbrinton/lcat',
      download_url='https://github.com/connorbrinton/lcat/archive/0.1.0.tar.gz',
      keywords=['CT', 'lung', 'biopsy'],
      license='GPL-3.0',
      package_data=PACKAGE_DATA,
      package_dir=PACKAGE_DIR,
      scripts=SCRIPTS)



"""
Utility functions for the lcat toolkit.
"""
import itertools
import os

import matplotlib.pyplot as plt
import numpy as np
import scipy



def plot_slices(voxels, rows=6, columns=6, cmap=None):
    """
    Given a 3D array of voxels, plots slices.
    """
    # Create subplots
    figure, axes = plt.subplots(rows, columns, figsize=[13, 13])

    figure.subplots_adjust(left=0.05, right=0.95, bottom=0.05, top=0.95)

    # List each figure location
    product = list(itertools.product(range(rows), range(columns)))

    # Create appropriate slice indices
    indices = list(np.linspace(0, voxels.shape[-1] - 1, len(product), dtype=int).flat)

    # Loop over figures
    for index, (row, column) in zip(indices, product):
        # Get specific subaxes
        subaxes = axes[row, column]

        # Set title
        subaxes.set_title("Slice %d" % index)

        # Display slice
        subaxes.imshow(voxels[:, :, index], cmap=cmap, interpolation='none')

        # Remove axes
        subaxes.axis('off')

    # Show plot
    plt.show()


def save_slices(voxels, destination_folder, prefix="slice"):
    """
    Given a voxel array, save each slice as a tiff file with the name `prefix###.tiff` in the given
    `destination_folder`.
    """
    # Make the directory if necessary
    try:
        os.mkdir(destination_folder)
    except OSError:
        pass

    # Determine number of digits in largest slice number
    width = len(str(voxels.shape[-1]))

    # Generate format string
    filename_template = "%s%%0%dd.tiff" % (prefix, width)

    # Save each slice
    for z_index in range(voxels.shape[-1]):
        destination_filepath = os.path.join(destination_folder, filename_template % (z_index))
        scipy.misc.imsave(destination_filepath, voxels[..., z_index])


def get_bounding_box(arr):
    """
    Given an array of values, returns an list of tuples, where each tuple represents the extent of
    the non-zero values in `arr` along a particular axis.

    Inspired by http://stackoverflow.com/questions/31400769/bounding-box-of-numpy-array
    """
    # Boundaries placeholder
    boundaries = []

    # For each axis
    for axis in range(arr.ndim):
        # Enumerate orthogonal axes
        other_axes = tuple(other_axis for other_axis in range(arr.ndim) if other_axis != axis)

        # Identify non-zero points
        nonzero = np.any(arr, axis=other_axes)

        # Identify non-zero region
        extent = tuple(np.where(nonzero)[0][[0, -1]])

        # Store boundary
        boundaries.append(extent)

    return boundaries


def crop_to_box(arr, box):
    """
    Given an array `arr` and boundaries `boundaries` (such as is returned by `get_bounding_box`),
    return `arr` cropped to `boundaries`.
    """
    # Calculate slices
    slicer = tuple(slice(extent[0], extent[1] + 1) for extent in box)

    # Perform slicing
    return arr[slicer]


def get_full_nodule_mask(nodule, scan_shape):
    # Create full placeholder
    mask = np.zeros(scan_shape, dtype=bool)

    # Create slicer
    slicer = tuple(slice(start, start + dim)
                   for start, dim in zip(nodule.origin, nodule.mask.shape))

    # Fill mask placeholder
    mask[slicer] = nodule.mask

    return mask


def compress_nodule_mask(mask):
    # Get bounding box
    bounding_box = get_bounding_box(mask)

    # Get origin from bounding box
    origin = [start for start, end in bounding_box]

    # Crop to bounding box
    mask = crop_to_box(mask, bounding_box)

    return origin, mask


def clear_border(labels, axis=None, in_place=False):
    """
    Clears any labeled components touching either border along the given axes (in `axis`).
    """
    # Convert labels to ndarray
    labels = np.asarray(labels)

    # Provide default axis argument
    if axis is None:
        axis = list(range(labels.ndim))

    # Make sure axis is 1d
    axes = np.atleast_1d(axis)
    assert axes.ndim == 1

    # Initialize to-be-cleared list
    touching = set()

    # For each axis
    for axis in axes.flat:
        # Generate selector
        selector = labels.ndim * [slice(None)]

        # Check leading border
        selector[axis] = 0
        touching |= set(np.unique(labels[tuple(selector)]))

        # Check trailing border
        selector[axis] = -1
        touching |= set(np.unique(labels[tuple(selector)]))

    if not in_place:
        labels = labels.copy()

    # Set matching values to zero
    for value in touching:
        labels[labels == value] = 0

    return labels


def image_from_mask(mask):
    """
    Convert a binary mask into a PIL image.
    """
    # Import PIL (if necessary)
    import PIL

    # Create mask image
    mask_image = PIL.Image.new('1', mask.shape)

    # Reserve memory for pixels
    pixels = mask_image.load()

    # Store pixel values
    for i in range(mask_image.size[0]):
        for j in range(mask_image.size[1]):
            pixels[i, j] = int(mask[i, j])

    return mask_image
